<?php
namespace Estoque\Logic\AtualizacaoEstoque;

use \Geral\Entities\Tipooperac;
use \Estoque\Entities\Itlanctoest;
use \Estoque\Entities\Lanctoestoq;
use \Estoque\Entities\Movilote;
use \Estoque\Logic\AtualizacaoEstoque\Exceptions\ProdutoEstoqueIrregular;
use \Geral\Util\DateUtil;
use Setforms2\Shared\Registry;
use \Setforms2\Shared\ORM\DateTime;
use \Setforms2\Runtime\Helper\Procedure;
use \Setforms2\Runtime\Helper\Procedure\Param;

/**
 * Created by JetBrains PhpStorm.
 * User: paulopereira
 * Date: 01/11/12
 * Time: 14:25
 * To change this template use File | Settings | File Templates.
 */
class AtualizadorDeEstoque
{

    /** @var $produtosIrregulares array */
    private static $produtosIrregulares;

    /**
     * Metodo responsavel por atualizar o estoque.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq  Entidade de Lancamento de estoque.
     * @param \Estoque\Entities\Itlanctoest $itlanctoest  Entidade de $itlanctoest, usada somente para atualizacao de
     *                                                    estoque por produto.
     * @param bool                          $lancaExcecao Verifica se o metodo deve ou nao lancar excecao, no caso de
     *                                                    produtos de estoque irregular.
     */
    public static function atualizarEstoque(Lanctoestoq $lanctoestoq, Itlanctoest $itlanctoest = null, $lancaExcecao = true)
    {
        // Altera o id de evolucao de estoque do lancamento ou do item.
        if (isset($itlanctoest)) {
            $itlanctoest->atualizaEvolucaoDeEstoque(Itlanctoest::ESTOQUE_EVOLUIDO);
            Registry::getEntityManager()->flush();
        } else {
            self::atualizaEvolucaoEstoquePorLancamento($lanctoestoq, Itlanctoest::ESTOQUE_EVOLUIDO);
        }

        // Executa a procedure de atualizacao de estoque.
        $dtfinal   = self::getDataFinal($lanctoestoq);
        $dtinicial = self::getDataInicial($lanctoestoq);
        self::executarProcedureAtualizaEstoque (
            $lanctoestoq, $dtfinal->format(DateUtil::FORMATO_DATA), $dtinicial->format(DateUtil::FORMATO_DATA)
        );

        // Executa o tratamento de produtos irregulares.
        self::tratamentoProdutosIrregulares($lanctoestoq, $dtfinal, $dtinicial, $lancaExcecao);
    }

    /**
     * Retorna a data final para atualizacao de estoque.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @return \DateTime
     */
    public static function getDataFinal(Lanctoestoq $lanctoestoq)
    {
        return self::getData($lanctoestoq->getDtlancestq(), DateUtil::DEPOIS);
    }

    /**
     * Retorna todos os itens que, apos a atualizacao de estoque permaneceram irregulares.
     *
     * @static
     * @return array
     */
    public static function getProdutosEstoqueIrregular()
    {
        return self::$produtosIrregulares;
    }

    /**
     * Retorna a data inicial de estoque, que devera ser a menor data entre a data de lancamento de estoque e a data
     * de processamento.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @return \DateTime|string
     */
    protected static function getDataInicial(Lanctoestoq $lanctoestoq)
    {
        return self::getData($lanctoestoq->getDtlancestq());
    }

    /**
     * Retorna as datas para estoque.
     *
     * @param \Setforms2\Shared\ORM\DateTime $dtinicial
     * @param int $modoComparacao
     * @return \DateTime|string
     */
    private function getData(\Setforms2\Shared\ORM\DateTime $dtinicial, $modoComparacao = DateUtil::ANTES)
    {
        $dtprocessamento = \Estoque\Util\Parametros::getDataProcessamento();
        $dtprocessamento = DateUtil::getDataDeString($dtprocessamento);

        // Retorna a data "mais tarde" entre as datas inicial e de processamento.
        return DateUtil::comparaDatas($dtinicial, $dtprocessamento, $modoComparacao) ? $dtinicial : $dtprocessamento;
    }

    /**
     * Atualiza o id de evolucao de todos os itens da evolucao de estoque passada por parametro.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @param $idEvolucao
     */
    private static function atualizaEvolucaoEstoquePorLancamento(Lanctoestoq $lanctoestoq, $idEvolucao)
    {
        /** @var $repositorio \Estoque\Repositories\Itlanctoest */
        $repositorio = Itlanctoest::getRepository();
        $repositorio->evoluiEstoquePorLancamento(
            $lanctoestoq->getFilial()->getCdfilial(), $lanctoestoq->getNrlancestq(),
            $lanctoestoq->getNrorg(), $idEvolucao
        );
    }

    /**
     * Executa procedure de Atualizacao de estoque (procedure ATUALIZAR_ESTOQUE).
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @param $dtfinal
     * @param $dtinicial
     * @throws \Exception
     */
    public static function executarProcedureAtualizaEstoque(Lanctoestoq $lanctoestoq, $dtfinal, $dtinicial)
    {
        try {
            $procedure = new Procedure("ATUALIZA_ESTOQUE");
            self::addParametrosAtualizaEstoque($procedure, $lanctoestoq, $dtfinal, $dtinicial);
            $procedure->call();
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Adiciona os parametros da procedure ATUALIZAR_ESTOQUE.
     *
     * @static
     * @param \Setforms2\Runtime\Helper\Procedure $procedure
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @param $dtfinal
     * @param $dtinicial
     */
    protected static function addParametrosAtualizaEstoque(Procedure $procedure, Lanctoestoq $lanctoestoq,
                                                           $dtfinal, $dtinicial)
    {
        $cdfilimovi   = $lanctoestoq->getFilimovi()->getCdfilial();
        $cdfilial     = $lanctoestoq->getFilial()->getCdfilial();
        $nrlancamento = $lanctoestoq->getNrlancestq();

        $procedure->addParam(
            new Param(
                'P_FILIMOVI',
                Param::PARAM_INPUT,
                $cdfilimovi,
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param(
                'P_DTINICIAL',
                Param::PARAM_INPUT,
                $dtinicial,
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param(
                'P_DTFINAL',
                Param::PARAM_INPUT,
                $dtfinal,
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param(
                'P_FILILANCTO',
                Param::PARAM_INPUT,
                $cdfilial,
                Param::PARAM_TYPE_STR
            )
        );
        $procedure->addParam(
            new Param(
                'P_NRLANCTO',
                Param::PARAM_INPUT,
                $nrlancamento,
                Param::PARAM_TYPE_STR
            )
        );

    }

    /**
     * Tratamento que visa regularizar os itens ate entao irregulares de estoque. Caso o metodo permita ()
     *
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @param \Datetime $dtfinal
     * @param $lancaExcecao
     * @param \Datetime $dtinicial
     *
     * @throws \Exception|ProdutoEstoqueIrregular
     */
    public static function tratamentoProdutosIrregulares(Lanctoestoq $lanctoestoq, $dtfinal, $dtinicial, $lancaExcecao)
    {
        try {
            $produtosEstoqueIrregular = self::getProdutosEstoqueIrregular($lanctoestoq);
            if ($produtosEstoqueIrregular) {
                /** @var $produtoIrregular Movilote */
                /** @var $itemAtualizar Itlanctoest */
                foreach ($produtosEstoqueIrregular as $produtoIrregular) {
                    $itemAtualizar = Itlanctoest::findOneByCdfilialAndNrlancestqAndCdprodutoAndNrorg(
                        $lanctoestoq->getFilial()->getCdfilial(),
                        $lanctoestoq->getNrlancestq(),
                        $produtoIrregular->getProduto()->getCdproduto(),
                        \Geral\Util\Ambiente::getNrorg()
                    );
                    $itemAtualizar->atualizaEvolucaoDeEstoque(Itlanctoest::ESTOQUE_NAO_EVOLUIDO);
                }

                self::executarProcedureAtualizaEstoque (
                    $lanctoestoq, $dtfinal->format(DateUtil::FORMATO_DATA), $dtinicial->format(DateUtil::FORMATO_DATA)
                );
                self::$produtosIrregulares = self::getProdutosMovilote($lanctoestoq);
                self::excluirProdutosMovilote($lanctoestoq);

                if ($lancaExcecao) {
                    throw new ProdutoEstoqueIrregular($produtosEstoqueIrregular);
                }
            }
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Retorna os produtos irregulares do movimento de estoque.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @return array
     */
    private  static function getProdutosMovilote(Lanctoestoq $lanctoestoq)
    {
        $nrmovilote = $lanctoestoq->getFilial()->getCdfilial().$lanctoestoq->getNrlancestq();
        $nrorg      = \Geral\Util\Ambiente::getNrorg();

        $produtosEstoqueIrregular = Movilote::findByNrmoviloteAndNrorg($nrmovilote, $nrorg);
        return $produtosEstoqueIrregular;
    }

    /**
     * Exclui todos os itens da tabeala de MOVILOTE (itens que nao puderam ser atualizados em estoque).
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     *
     * @throws \Exception
     */
    private static function excluirProdutosMovilote(Lanctoestoq $lanctoestoq)
    {
        try {
            $produtos = self::getProdutosMovilote($lanctoestoq);
            /** @var $produto Movilote*/
            foreach ($produtos as $produto) {
                $produto->delete();
            }

            \Setforms2\Shared\Registry::getEntityManager()->flush();
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * Verifica se o tipo de operacao da nota fiscal exige atualizacao de estoque.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lanctoestoq
     * @return bool
     */
    public static function autorizaAtualizacaoDeEstoque(Lanctoestoq $lanctoestoq)
    {
        $idatuesttpop = $lanctoestoq->getTipooper()->getIdatuesttpop();
        return $idatuesttpop === Tipooperac::AUTORIZA_ATUALIZACAO_ESTOQUE;
    }

    /**
     * Verifica se houve lancamento de estoque para a nota fiscal de origem.
     *
     * @param \Fiscal\Entities\Notafiscal $notaFiscal
     * @throws \Estoque\Logic\AtualizacaoEstoque\Exceptions\LancamentoNaoAtualizado
     */
    public static function verificarLancamentoEstoqueDaNota(\Fiscal\Entities\Notafiscal $notaFiscal)
    {
        $cdintegracao = $notaFiscal->getCdintegracao();
        $idatuesttpop = $notaFiscal->getEntrsaidop()->getIdatuesttpop();
        // Verifica se o campo de cdintegracao nao foi preenchido e se o tipo de operacao da nota fiscal permite
        // atualizacao de estoque.
        if ((bool)$cdintegracao === false && $idatuesttpop === Tipooperac::AUTORIZA_ATUALIZACAO_ESTOQUE) {
            $lancamento = Lanctoestoq::findOneByCdfilialAndNrlanctonfAndNrorg (
                $notaFiscal->getFilial()->getCdfilial(),
                $notaFiscal->getNrlanctonf(),
                \Geral\Util\Ambiente::getNrorg()
            );
            if ((bool)$lancamento === false) {
                throw new \Estoque\Logic\AtualizacaoEstoque\Exceptions\LancamentoNaoAtualizado();
            }
        }

    }

    /**
     * Atualiza o operador do lancamento de estoque.
     *
     * @static
     * @param \Estoque\Entities\Lanctoestoq $lancamento
     */
    public static function atualizaOperadorLancamentoEstoque(Lanctoestoq $lancamento)
    {
        /* @var $operador \Geral\Entities\Operador */
        $operador = \Geral\Entities\Operador::getObject(
            array(
                "cdoperador" => \Geral\Util\Ambiente::getCdOperador(),
                "nrorg"      => \Geral\Util\Ambiente::getNrorg(),
            )
        );
        $lancamento->setOperador($operador);
        $lancamento->persist();
    }


}
